@@ -7,6 +7,7 @@
 --
 -- or, if you're picky, you can invoke only 'deriveFunctor' and whichever other instances you need instead.
 
+{-# Language CPP #-}
 {-# Language TemplateHaskell #-}
 -- Adapted from https://wiki.haskell.org/A_practical_Template_Haskell_Tutorial
 
@@ -46,6 +47,7 @@ deriveApply ty = do
               pragInlD '(Rank2.<*>) Inlinable FunLike AllPhases,
               pragInlD 'Rank2.liftA2 Inlinable FunLike AllPhases]]
 
+-- | This function always succeeds, but the methods it generates may be partial. Use with care.
 unsafeDeriveApply :: Name -> Q [Dec]
 unsafeDeriveApply ty = do
    (instanceType, cs) <- reifyConstructors ''Rank2.Apply ty
@@ -97,10 +99,17 @@ reifyConstructors cls ty = do
       NewtypeD _ nm tyVars kind c _ -> return (nm, tyVars, kind, [c])
       _ -> fail "deriveApply: tyCon may not be a type synonym."
  
-   let (KindedTV tyVar (AppT (AppT ArrowT StarT) StarT)) = last tyVars
+#if MIN_VERSION_template_haskell(2,17,0)
+   let (KindedTV tyVar () (AppT (AppT ArrowT _) StarT)) = last tyVars
+       instanceType           = conT cls `appT` foldl apply (conT tyConName) (init tyVars)
+       apply t (PlainTV name _)    = appT t (varT name)
+       apply t (KindedTV name _ _) = appT t (varT name)
+#else
+   let (KindedTV tyVar (AppT (AppT ArrowT _) StarT)) = last tyVars
        instanceType           = conT cls `appT` foldl apply (conT tyConName) (init tyVars)
        apply t (PlainTV name)    = appT t (varT name)
        apply t (KindedTV name _) = appT t (varT name)
+#endif
  
    putQ (Deriving tyConName tyVar)
    return (instanceType, cs)
@@ -171,7 +180,7 @@ genFmapClause (RecC name fields) = do
           ((,) fieldName <$>)
           <$> genFmapField (varE f) fieldType (appE (varE fieldName) (varE x)) id
    constraints <- (concat . (fst <$>)) <$> sequence constraintsAndFields
-   (,) constraints <$> clause [varP f, bangP (varP x)] body []
+   (,) constraints <$> clause [varP f, x `asP` recP name []] body []
 genFmapClause (GadtC [name] fieldTypes _resultType@(AppT _ (VarT tyVar))) =
    do Just (Deriving tyConName _tyVar) <- getQ
       putQ (Deriving tyConName tyVar)
@@ -213,7 +222,7 @@ genLiftA2Clause unsafely (RecC name fields) = do
        newNamedField (fieldName, _, fieldType) =
           fieldExp fieldName (genLiftA2Field unsafely (varE f) fieldType (getFieldOf x) (getFieldOf y) id)
           where getFieldOf = appE (varE fieldName) . varE
-   clause [varP f, bangP (varP x), varP y] body []
+   clause [varP f, x `asP` recP name [], varP y] body []
 genLiftA2Clause unsafely (GadtC [name] fieldTypes _resultType@(AppT _ (VarT tyVar))) =
    do Just (Deriving tyConName _tyVar) <- getQ
       putQ (Deriving tyConName tyVar)
@@ -234,7 +243,7 @@ genLiftA2Field unsafely fun fieldType field1Access field2Access wrap = do
         | t1 /= VarT typeVar -> genLiftA2Field unsafely fun t2 field1Access field2Access (appE (varE 'liftA2) . wrap)
      SigT ty _kind -> genLiftA2Field unsafely fun ty field1Access field2Access wrap
      ParensT ty -> genLiftA2Field unsafely fun ty field1Access field2Access wrap
-     _ | unsafely -> [| error "Cannot apply liftA2 to field" |]
+     _ | unsafely -> field1Access
        | otherwise -> error ("Cannot apply liftA2 to field of type " <> show fieldType)
 
 genLiftA3Clause :: Bool -> Con -> Q Clause
@@ -261,7 +270,7 @@ genLiftA3Clause unsafely (RecC name fields) = do
        newNamedField (fieldName, _, fieldType) =
           fieldExp fieldName (genLiftA3Field unsafely (varE f) fieldType (getFieldOf x) (getFieldOf y) (getFieldOf z) id)
           where getFieldOf = appE (varE fieldName) . varE
-   clause [varP f, bangP (varP x), varP y, varP z] body []
+   clause [varP f, x `asP` recP name [], varP y, varP z] body []
 genLiftA3Clause unsafely (GadtC [name] fieldTypes _resultType@(AppT _ (VarT tyVar))) =
    do Just (Deriving tyConName _tyVar) <- getQ
       putQ (Deriving tyConName tyVar)
@@ -285,7 +294,7 @@ genLiftA3Field unsafely fun fieldType field1Access field2Access field3Access wra
           -> genLiftA3Field unsafely fun t2 field1Access field2Access field3Access (appE (varE 'liftA3) . wrap)
      SigT ty _kind -> genLiftA3Field unsafely fun ty field1Access field2Access field3Access wrap
      ParensT ty -> genLiftA3Field unsafely fun ty field1Access field2Access field3Access wrap
-     _ | unsafely -> [| error "Cannot apply liftA3 to field" |]
+     _ | unsafely -> field1Access
        | otherwise -> error ("Cannot apply liftA3 to field of type " <> show fieldType)
 
 genApClause :: Bool -> Con -> Q ([Type], Clause)
@@ -311,7 +320,7 @@ genApClause unsafely (RecC name fields) = do
           ((,) fieldName <$>) <$> genApField unsafely fieldType (getFieldOf x) (getFieldOf y) id
           where getFieldOf = appE (varE fieldName) . varE
    constraints <- (concat . (fst <$>)) <$> sequence constraintsAndFields
-   (,) constraints <$> clause [varP x, bangP (varP y)] body []
+   (,) constraints <$> clause [x `asP` recP name [], varP y] body []
 genApClause unsafely (GadtC [name] fieldTypes _resultType@(AppT _ (VarT tyVar))) =
    do Just (Deriving tyConName _tyVar) <- getQ
       putQ (Deriving tyConName tyVar)
@@ -332,7 +341,7 @@ genApField unsafely fieldType field1Access field2Access wrap = do
      AppT t1 t2 | t1 /= VarT typeVar -> genApField unsafely t2 field1Access field2Access (appE (varE 'liftA2) . wrap)
      SigT ty _kind -> genApField unsafely ty field1Access field2Access wrap
      ParensT ty -> genApField unsafely ty field1Access field2Access wrap
-     _ | unsafely -> (,) [] <$> [| error ("Cannot apply ap to field" <> $(pure $ LitE $ StringL $ show fieldType)) |]
+     _ | unsafely -> (,) [] <$> field1Access
        | otherwise -> error ("Cannot apply ap to field of type " <> show fieldType)
 
 genPureClause :: Con -> Q ([Type], Clause)
@@ -377,7 +386,7 @@ genFoldMapClause (NormalC name fieldTypes) = do
        newField x (_, fieldType) = genFoldMapField f fieldType (varE x) id
    constraints <- (concat . (fst <$>)) <$> sequence constraintsAndFields
    (,) constraints <$> clause pats (normalB body) []
-genFoldMapClause (RecC _name fields) = do
+genFoldMapClause (RecC name fields) = do
    f <- newName "f"
    x <- newName "x"
    let body | null fields = [| mempty |]
@@ -387,7 +396,7 @@ genFoldMapClause (RecC _name fields) = do
        newField :: VarBangType -> Q ([Type], Exp)
        newField (fieldName, _, fieldType) = genFoldMapField f fieldType (appE (varE fieldName) (varE x)) id
    constraints <- (concat . (fst <$>)) <$> sequence constraintsAndFields
-   (,) constraints <$> clause [varP f, bangP (varP x)] (normalB body) []
+   (,) constraints <$> clause [varP f, x `asP` recP name []] (normalB body) []
 genFoldMapClause (GadtC [name] fieldTypes _resultType@(AppT _ (VarT tyVar))) =
    do Just (Deriving tyConName _tyVar) <- getQ
       putQ (Deriving tyConName tyVar)
@@ -412,7 +421,7 @@ genFoldMapField funcName fieldType fieldAccess wrap = do
 
 genTraverseClause :: Con -> Q ([Type], Clause)
 genTraverseClause (NormalC name []) =
-   (,) [] <$> clause [wildP, wildP] (normalB [| pure $(conE name) |]) []
+   (,) [] <$> clause [wildP, conP name []] (normalB [| pure $(conE name) |]) []
 genTraverseClause (NormalC name fieldTypes) = do
    f          <- newName "f"
    fieldNames <- replicateM (length fieldTypes) (newName "x")
@@ -436,7 +445,7 @@ genTraverseClause (RecC name fields) = do
        newField :: VarBangType -> Q ([Type], Exp)
        newField (fieldName, _, fieldType) = genTraverseField (varE f) fieldType (appE (varE fieldName) (varE x)) id
    constraints <- (concat . (fst <$>)) <$> sequence constraintsAndFields
-   (,) constraints <$> clause [varP f, bangP (varP x)] body []
+   (,) constraints <$> clause [varP f, x `asP` recP name []] body []
 genTraverseClause (GadtC [name] fieldTypes _resultType@(AppT _ (VarT tyVar))) =
    do Just (Deriving tyConName _tyVar) <- getQ
       putQ (Deriving tyConName tyVar)
